/* ****************************************** */
/* Emu75: HP-75  software emulator for PC     */
/*                                            */
/* capric1.c   Capricorn emulation module - 1 */
/*                                            */
/* Version 1.1  J-F Garnier 2010-2017         */
/* ****************************************** */


//#include <mem.h>
//#include <stdio.h>
#include <dos.h>


#define GLOBAL extern
#include "capricor.h"

extern void disp_string(char *s);    // in userpc.c

/* macro to build a pointer into Emu75 memory space. */
#define MK_FP75(a) MK_FP(tab_seg[a>>13],(int)(a&0x1fff))


/* ****************************************** */
/* int fetch1(unsigned int adr)               */
/*                                            */
/* fetch one opcode from address addr         */
/* ****************************************** */
unsigned char fetch1(unsigned int addr)
{
  int p0, n;
  static unsigned char far *ptr=0;
  static unsigned int lastaddr=0xffff;

  /* lastaddr holds the last addr + 1 to avoid pointer calculations
     when fetching successive words */
  if (addr!=lastaddr) {
    lastaddr=addr;
    p0=addr>>13; /* page number*/
    addr &=0x1fff;
    if (tab_seg[p0]==0) {
      /* nonexistent page */
      lastaddr=0xffff;
      return(0);
    }
    /* setup pointer */
    ptr= (unsigned char far *) MK_FP(tab_seg[p0],addr);
  }

  n=*ptr++;
  lastaddr++;
  return(n);
}

/* ****************************************** */
/* int storeData(int adr)                     */
/*                                            */
/* store data to RAM or I/O                   */
/* ****************************************** */
int storeData(unsigned int addr, unsigned char *data, int n)
{
  unsigned char far *ptr;
  int res;
  int powerstat, p;
  char ligne[40];

  res=0;
  if ((addr>=0x8000)&&(addr<0x0e000)) {
    /* RAM */
    ptr=MK_FP(tab_seg[4], addr-0x8000);
    while (n--) *ptr++ = *data++;
  }
  else if (addr<0xff00) {
    /* ROM, ignore unless PMS */
    if ((addr>=0x6000)&&(addr<0x8000)) {
      // plug in module
//    sprintf(ligne,"ROM write attempt at slot %02x\r\n",regROMslt);
//    disp_string(ligne);
      if (fwrom) {
        // if writeable (PMS);
        if (tab_seg[3]!=0) {
          ptr=MK_FP(tab_seg[3], addr&0x1fff);
          while (n--) *ptr++ = *data++;
//    sprintf(ligne,"ROM write at slot %02x\r\n",regROMslt);
//    disp_string(ligne);
        }
      }
    }
  }
  else {
    /* I/O */
//    sprintf(ligne,"I/O write at %04X, %d bytes\n",addr,n);
//    disp_string(ligne);
    if ((addr>=0xff30)&(addr<0xff60)) {
      // ROM switch
      if ((addr&1)==0) {
        // try to select a ROM
        selectbank(addr&0xff);
        regROMslt=addr&0xff;
      }
      else
        if ((addr>=0xff40)&(addr<0xff50)) {
          // disable ROM - only ff4x v1.1
          selectbank(0);
          regROMslt=0;
        }
    }
    else {

      ptr= (unsigned char far *) MK_FP(tab_seg[7],(addr&0x1fff));

      switch (addr) {

          // global interrupt enable/disable
        case 0xff00:  fintr=1;  break;
        case 0xff01:  fintr=0;  break;

        case 0xff02:
          // key status
          // bit 1: enable IRQ, also reset pending IRQ
          // bit 5: physical key available
          *ptr = data[0];
          break;

        case 0xff08:
          // card reader
          *ptr = data[0];
          break;

        case 0xff10:
        case 0xff11:
        case 0xff12:
        case 0xff13:
        case 0xff14:
        case 0xff15:
        case 0xff16:
        case 0xff17:
          // hpil - contrary to other I/Os, HPIL registers can be accessed in multibyte ops
          while (n--) {
//    sprintf(ligne," write %02x to reg %x    ",*data,addr&7);
//    disp_string(ligne);
            *ptr++ = *data;  // memory-mapped copy
            hpil_wr(addr&7, *data);
            addr++;
            data++;
          }
//    disp_string("\r\n");
//        if (n!=1) res=-1;   // invalid?
          break;

        case 0xff80:
          //  comparator
          // first byte =status. bit1= enable IRQ
          *ptr = data[0];
          compstat=data[0];
          // next 5 bytes are next timer IRQ
          if (n>=6) {
            tcomp_l=data[1];   // low part
            tcomp = data[5];
            tcomp<<=8;;
            tcomp += data[4];
            tcomp<<=8;;
            tcomp += data[3];
            tcomp<<=8;;
            tcomp += data[2];
          }
          break;

        case 0xff81:
          //  timer
          // first byte =status.
          *ptr = data[0];
          // next 5 bytes are timer value (counting up)
          if (n>=6) {
            timer_l=data[1];   // low part
            timer = data[5];
            timer<<=8;;
            timer += data[4];
            timer<<=8;;
            timer += data[3];
            timer<<=8;;
            timer += data[2];
          }
          break;

        case 0xff82:
          // power status
          powerstat=data[0];
          *ptr = powerstat;
          // light sleep?
          if (powerstat&0x40)
            if (powerstat&0x20)
              res=1; // light sleep
            else
              res=2; // deep sleep
          break;

        case 0xfffc:
          // lcd
          // 1st byte is lcd position
          p=data[0];
          if (n>1) {
            // 2nd byte is lcd data
            if (p==0) {
              // write data in all LCD positions
              for (p=0;p<34;p++) regLCD[p]=data[1];
            }
            else if ((p>=47)&&(p<=80)) {
              // write data in LCD position, 80 is left, 49 right, 48-47 are announciators
              regLCD[80-p]=data[1];
            }
          }
          break;

        default:
          // for test: write one byte
//          *ptr = data[0];
//          res=-1;
            ;
      } // end switch addr

    } // end reg I/O
  } // end I/O write

  return(res);
}


/* ****************************************** */
/* int loadData(int adr)                      */
/*                                            */
/* load data from ROM, RAM or I/O             */
/* ****************************************** */
int readData(unsigned int addr, unsigned char *data, int n)
{
  unsigned char far *ptr;
  int p0;
  char ligne[40];
  int res;

  res=0;
  if ((addr>=0x8000)&&(addr<0x0e000)) {
    /* RAM */
    ptr=MK_FP(tab_seg[4], addr-0x8000);
    while (n--) *data++ = *ptr++;
  }
  else if (addr<0xff00) {
    /* ROM */
    p0=addr>>13;
    addr &=0x1fff;
    if (tab_seg[p0]==0) {
      /* nonexistent page */
      while (n--) *data++ = 0xff;
    }
    else {
      /* setup pointer */
      ptr= (unsigned char far *) MK_FP(tab_seg[p0],addr);
      while (n--) *data++ = *ptr++;
    }
  }
  else {
    /* I/O */
//    sprintf(ligne,"I/O read at %04X, %d bytes\n",addr,n);
//    disp_string(ligne);

    ptr= (unsigned char far *) MK_FP(tab_seg[7],(addr&0x1fff));

    switch (addr) {

      case 0xff00:
      case 0xff01:
        data[0]= *ptr;
        break;

      case 0xff02:
        data[0]= *ptr;   // key status
        // v1.1: manage the physical key register for diag ROM
        if (n>1)
          data[1]=regKey;
        *ptr &= 0xdf; // reset key avail (is it the right way to do?)
        break;

      case 0xff08:
        break;

      case 0xff10:
      case 0xff11:
      case 0xff12:
      case 0xff13:
      case 0xff14:
      case 0xff15:
      case 0xff16:
      case 0xff17:
        // hpil - contrary to other I/Os, HPIL can be accessed in multibyte ops
        while (n--) {
          *data=hpil_rd(addr&7);
          *ptr++ = *data;  // memory-mapped copy
//      sprintf(ligne,"read %02x from reg %x   ",*data,addr&7);
//      disp_string(ligne);
          addr++;
          data++;
         }
//      disp_string("\r\n");
//        if (n!=1) res=-1;   // invalid?
        break;

      case 0xff80:
        //  comparator
        // first byte =status. bit1= enable IRQ
        data[0] = *ptr;
        // next 5 bytes are next timer IRQ
        if (n>=6) {
          data[1] = tcomp_l;   // low part
          data[5] = tcomp>>24;
          data[4] = tcomp>>16;
          data[3] = tcomp>>8;
          data[2] = tcomp;
        }
        break;

      case 0xff81:
        //  timer
        // first byte =status.
        data[0] = *ptr;
        // next 5 bytes are timer value (counting up)
        if (n>=6) {
          data[1]=timer_l;   // low part
          data[5]=timer>>24;
          data[4]=timer>>16;
          data[3]=timer>>8;
          data[2]=timer;
        }
        break;

      case 0xff82:
        // power status
        data[0] = *ptr;
        break;

      case 0xfffc:
        //  LCD
        // first byte =status. LCD ready if bit0=0
        data[0] = (*ptr) & 0xfe;  // always ready
        break;

      default:
        // for test: read 1 byte
        *data++ = *ptr; n--;
        while (n--) *data++ = 0;

    } // end switch addr

  }

  return(res);
}



// do interrupt
void do_intr(int vector)
{
  int v;
  unsigned char buf[2];

  // push PC
  v=regCPU.reg.pc ;
  buf[0]=v; buf[1]=v>>8;
  storeData(regCPU.reg.sp,buf,2);
  // SP +=2
  regCPU.reg.sp += 2;

 // jump
  v=lit2(vector);
  regCPU.reg.pc = v;

}



// EA calculations


// literal immediat
void lit_imm(unsigned char *sptr, int n)
{
  int v;

  regCPU.reg.pc++;
  v=regCPU.reg.pc;
  *sptr++ = fetch1(v++);
  regCPU.reg.pc += n;
  while (n--) {
    *sptr++ = fetch1(v++);
  }
}

// literal direct
int lit_direct(void)
{
  unsigned int v;

  regCPU.reg.pc++;
  v=fetch1(regCPU.reg.pc);
  regCPU.reg.pc++;
  v += fetch1(regCPU.reg.pc)<<8;
  return(v);
}

// index direct
int index_direct(void)
{
  unsigned int v;

  regCPU.reg.pc++;
  v=fetch1(regCPU.reg.pc);
  regCPU.reg.pc++;
  v += fetch1(regCPU.reg.pc)<<8;
  v += regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
//  regCPU.reg.idx= v;  // index stored in R02 - not done for jsb  26/02/17
  return(v);
}

// register indirect
int reg_ind(void)
{
  unsigned int v;
  unsigned char buf[2];

  v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
  readData(v, buf, 2);
  v=buf[0]+(buf[1]<<8);
  return(v);
}

// litteral indirect
int lit_ind(void)
{
  unsigned int v;
  unsigned char buf[2];

  regCPU.reg.pc++;
  v=fetch1(regCPU.reg.pc);
  regCPU.reg.pc++;
  v += fetch1(regCPU.reg.pc)<<8;
  readData(v, buf, 2);
  v=buf[0]+(buf[1]<<8);
  return(v);
}

// index indirect
int index_ind(void)
{
  unsigned int v;
  unsigned char buf[2];

  regCPU.reg.pc++;
  v=fetch1(regCPU.reg.pc);
  regCPU.reg.pc++;
  v += fetch1(regCPU.reg.pc)<<8;
  v += regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
  regCPU.reg.idx= v;  // index stored in R02
  readData(v, buf, 2);
  v=buf[0]+(buf[1]<<8);
  return(v);
}



/* ****************************************** */
/* int executeCapricorn(int n)                */
/*                                            */
/* executes n Capricorn opcodes               */
/* returns:  0 if OK,                         */
/*   -1 if invalid, -2 if break               */
/*    1 if light sleep, 2 if deep sleep       */
/* ****************************************** */
int executeCapricorn(int nb_instr)
{
  unsigned char opcode;
  int res;
  int tmpreg;
  int i, n, f;
  unsigned int v;
  unsigned char buf[16];
  unsigned char *dptr;  // dest pointer (usually DR)
  unsigned char *sptr;  // src pointer (usually pointed by AR)

  res=0;
  while (nb_instr--) {

    // fetch an opcode
    opcode=fetch1(regCPU.reg.pc);

    dptr= &regCPU.byt[regDRP];  // init pointer to DR - used by so many opcodes

    // decode
    switch (opcode) {

      // arp
      case 0x01:
        regARP=regCPU.byt[0]&0x3f;
        break;
      case 0x00:
      case 0x02:
      case 0x03:
      case 0x04:
      case 0x05:
      case 0x06:
      case 0x07:
      case 0x08:
      case 0x09:
      case 0x0a:
      case 0x0b:
      case 0x0c:
      case 0x0d:
      case 0x0e:
      case 0x0f:
      case 0x10:
      case 0x11:
      case 0x12:
      case 0x13:
      case 0x14:
      case 0x15:
      case 0x16:
      case 0x17:
      case 0x18:
      case 0x19:
      case 0x1a:
      case 0x1b:
      case 0x1c:
      case 0x1d:
      case 0x1e:
      case 0x1f:
      case 0x20:
      case 0x21:
      case 0x22:
      case 0x23:
      case 0x24:
      case 0x25:
      case 0x26:
      case 0x27:
      case 0x28:
      case 0x29:
      case 0x2a:
      case 0x2b:
      case 0x2c:
      case 0x2d:
      case 0x2e:
      case 0x2f:
      case 0x30:
      case 0x31:
      case 0x32:
      case 0x33:
      case 0x34:
      case 0x35:
      case 0x36:
      case 0x37:
      case 0x38:
      case 0x39:
      case 0x3a:
      case 0x3b:
      case 0x3c:
      case 0x3d:
      case 0x3e:
      case 0x3f:
        regARP=opcode&0x3f;
        break;

      // drp
      case 0x41:
        regDRP=regCPU.byt[0]&0x3f;
//        lenDR= (~regDRP)&7; if (!(regDRP&0x20)) lenDR &=1;
        if (regDRP&0x20)
          lenDR= 7- (regDRP&7);
        else
          lenDR= 1 - (regDRP&1);
        break;
      case 0x40:
      case 0x42:
      case 0x43:
      case 0x44:
      case 0x45:
      case 0x46:
      case 0x47:
      case 0x48:
      case 0x49:
      case 0x4a:
      case 0x4b:
      case 0x4c:
      case 0x4d:
      case 0x4e:
      case 0x4f:
      case 0x50:
      case 0x51:
      case 0x52:
      case 0x53:
      case 0x54:
      case 0x55:
      case 0x56:
      case 0x57:
      case 0x58:
      case 0x59:
      case 0x5a:
      case 0x5b:
      case 0x5c:
      case 0x5d:
      case 0x5e:
      case 0x5f:
      case 0x60:
      case 0x61:
      case 0x62:
      case 0x63:
      case 0x64:
      case 0x65:
      case 0x66:
      case 0x67:
      case 0x68:
      case 0x69:
      case 0x6a:
      case 0x6b:
      case 0x6c:
      case 0x6d:
      case 0x6e:
      case 0x6f:
      case 0x70:
      case 0x71:
      case 0x72:
      case 0x73:
      case 0x74:
      case 0x75:
      case 0x76:
      case 0x77:
      case 0x78:
      case 0x79:
      case 0x7a:
      case 0x7b:
      case 0x7c:
      case 0x7d:
      case 0x7e:
      case 0x7f:
        regDRP=opcode&0x3f;
        if (regDRP&0x20)
          lenDR= 7- (regDRP&7);
        else
          lenDR= 1 - (regDRP&1);
//        lenDR= (~regDRP)&7; if ((regDRP&0x20)==0) lenDR &=1;
        break;


      // --------------- shift  -----------------------

      case 0x84: // LLB
        // LL, LR  : init E or CY to 0
        if (regFlags & m_DCM) regE=0;
        regFlags &= ~m_CY;
        // fall into ELB
      case 0x80: // ELB
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)<<4)+regE;
          //  init flags
          regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
          regE=tmpreg>>8;
        }
        else {
          // binary mode
          tmpreg=(*dptr)<<1;
          if (regFlags&m_CY) tmpreg++;
          //  init flags
          regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
          if (tmpreg&0x100) regFlags |= m_CY;
          //  in BIN mode: OVF set if shift causes a sign change
          if ( (tmpreg&0x80)!=((*dptr)&0x80) ) regFlags |= m_OVF;
        }
endByte:
        // update Z, MSB, LSB, LDZ, RDZ
        regFlags |= tmpreg & (m_MSB+m_LSB);
        if ((tmpreg&0xFF)==0) regFlags |= m_Z;
        if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        // store byte
        *dptr=tmpreg;
        break;

      case 0x85: // LLM
        // LL, LR  : init E or CY to 0
        if (regFlags & m_DCM) regE=0;
        regFlags &= ~m_CY;
        // fall into ELM
      case 0x81: // ELM
        // shift left
        n=lenDR;
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)<<4) + regE;
          //  init flags
          regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
          regFlags |= m_Z;
          regE=tmpreg>>8;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            tmpreg=((*dptr)<<4) + regE;
            regE=tmpreg>>8;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
        }
        else {
          // binary mode
          tmpreg=(*dptr)<<1;
          if (regFlags&m_CY) tmpreg++;
          //  init flags
          regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
          regFlags |= m_Z;
          if (tmpreg&0x100) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            tmpreg=(*dptr)<<1;
            if (regFlags&m_CY) tmpreg++;
            if (tmpreg&0x100) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          };
          //  in BIN mode: OVF set if shift causes a sign change
          if ( (tmpreg&0x80)!=((*dptr)&0x80) ) regFlags |= m_OVF;
        }
endMultiple:
        // update MSB, LDZ
        regFlags |= tmpreg & (m_MSB);
        if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
        // store byte
        *dptr=tmpreg;
        break;



      case 0x86: // LRB
        // LL, LR  : init E or CY to 0
        if (regFlags & m_DCM) regE=0;
        regFlags &= ~m_CY;
        // fall into ERB
      case 0x82: // ERB
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=(regE<<8)+(*dptr);
          //  init flags
          regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
          regE=tmpreg&0x0F;
          tmpreg >>= 4;
        }
        else {
          // binary mode
          tmpreg= *dptr;
          if (regFlags&m_CY) tmpreg+=0x100;
          //  init flags
          regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
          if (tmpreg&0x1) regFlags |= m_CY;
          tmpreg >>= 1;
          //  in BIN mode: OVF set if shift causes a sign change
          // V1.1: no, OVF never set
//          if ( (tmpreg&0x80)!=((*dptr)&0x80) ) regFlags |= m_OVF;
        }
        goto endByte;


      case 0x87: // LRM
        // LL, LR  : init E or CY to 0
        if (regFlags & m_DCM) regE=0;
        regFlags &= ~m_CY;
        // fall into ERM
      case 0x83: // ERM
        // shift right
        // multibyte shift right proceeds from DR to preceding (lower-numbered) boundary
        if (regDRP&0x20)
          n= (regDRP&7);
        else
          n= (regDRP&1);
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=(regE<<8)+(*dptr);
          //  init flags
          regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
          regFlags |= m_Z;
          regE=tmpreg&0x0F;
          tmpreg >>= 4;
          // update flags for first byte (MSB)
          regFlags |= tmpreg & (m_MSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
          while (n--) {
            *dptr-- = tmpreg;
            tmpreg=(regE<<8)+(*dptr);
            regE=tmpreg&0x0F;
            tmpreg >>= 4;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
        }
        else {
          // binary mode
          tmpreg= *dptr;
          if (regFlags&m_CY) tmpreg+=0x100;
          //  init flags
          regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
          regFlags |= m_Z;
          if (tmpreg&0x1) regFlags |= m_CY;
          tmpreg >>= 1;
          // update flags for first byte (MSB)
          regFlags |= tmpreg & (m_MSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
          //  in BIN mode: OVF set if shift causes a sign change
          // V1.1: no, OVF never set
//          if ( (tmpreg&0x80)!=((*dptr)&0x80) ) regFlags |= m_OVF;
          while (n--) {
            *dptr-- = tmpreg; // towards preceding bytes
            tmpreg= *dptr;
            if (regFlags&m_CY) tmpreg+=0x100;
            if (tmpreg&0x1) regFlags |= m_CY; else regFlags &= ~m_CY;
            tmpreg >>= 1;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
        }
        // update LSB, RDZ
        regFlags |= tmpreg & (m_LSB);
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        // store byte
        *dptr =tmpreg;
        break;


      // --------------- incr/decr -----------------------

      case 0x88: // ICB
        //  init flags
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)+1;
          if (tmpreg>0x09) tmpreg += 6; // -10 + 16
          tmpreg += (*dptr)&0xF0;
          if (tmpreg>0x99) tmpreg += 0x60;  // -10*16
        }
        else {
          // binary mode
          tmpreg=(*dptr)+1;
          //  in BIN mode: OVF set if sign changes to negative
          if ( (tmpreg&0x80)&&(((*dptr)&0x80)==0) ) regFlags |= m_OVF;
        }
        if (tmpreg&0x100) regFlags |= m_CY;
        goto endByte;

      case 0x89: // ICM
        n=lenDR;
        //  init flags
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)+1;
          if (tmpreg>0x09) tmpreg += 6; // -10 + 16
          tmpreg += (*dptr)&0xF0;
          if (tmpreg>0x99) tmpreg += 0x60;  // -10*16 + 16*16
          if (tmpreg&0x100) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            tmpreg=((*dptr)&0x0F);
            if (regFlags & m_CY) tmpreg++;
            if (tmpreg>0x09) tmpreg += 6; // -10 + 16
            tmpreg += (*dptr)&0xF0;
            if (tmpreg>0x99) tmpreg += 0x60;  // -10*16
            if (tmpreg&0x100) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
        }
        else {
          // binary mode
          tmpreg=(*dptr)+1;
          if (tmpreg&0x100) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            tmpreg = *dptr;
            if (regFlags & m_CY) tmpreg++;
            if (tmpreg&0x100) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
          //  in BIN mode: OVF set if sign changes to negative
          if ( (tmpreg&0x80)&&(((*dptr)&0x80)==0) ) regFlags |= m_OVF;
        }
        goto endMultiple;


      case 0x8a: // DCB
        //  init flags
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)-1;
          if (tmpreg<0) tmpreg -= 6;  // +10 -16
          tmpreg += (*dptr)&0xF0;
          if (tmpreg<0) tmpreg -= 0x60;
        }
        else {
          // binary mode
          tmpreg=(*dptr)-1;
          //  in BIN mode: OVF set if sign changes to positive
          if ( ((tmpreg&0x80)==0)&&((*dptr)&0x80) ) regFlags |= m_OVF;
        }
        // update CY state: set if decrementing a non-zero number
        if ((*dptr)) regFlags |= m_CY;
        goto endByte;

      case 0x8b: // DCM
        n=lenDR;
        //  init flags
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)-1;
          if (tmpreg<0) tmpreg -= 6; // +10 - 16
          tmpreg += (*dptr)&0xF0;
          if (tmpreg<0) tmpreg -= 0x60;  // +10*16 - 16*16
          if (tmpreg<0) regFlags |= m_CY;  // tbc
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            tmpreg=((*dptr)&0x0F);
            if (regFlags & m_CY) tmpreg--;
            if (tmpreg<0) tmpreg -= 6; // -10 + 16
            tmpreg += (*dptr)&0xF0;
            if (tmpreg<0) tmpreg -= 0x60;  // -10*16
            if (tmpreg<0) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
        }
        else {
          // binary mode
          tmpreg=(*dptr)-1;
          if (tmpreg<0) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            tmpreg= *dptr;
            if (regFlags & m_CY) tmpreg--;
            if (tmpreg<0) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
          //  in BIN mode: OVF set if sign changes to positive
          if ( ((tmpreg&0x80)==0)&&((*dptr)&0x80) ) regFlags |= m_OVF;
        }
        // invert CY state: set if decrementing a non-zero number
        regFlags ^= m_CY;
        goto endMultiple;


      // --------------- complement -----------------------

      case 0x8c: // TCB
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        if (regFlags&m_DCM) {
          // decimal mode (do 100-content)
          tmpreg= 10 - ((*dptr)&0x0F);
          if (tmpreg>9) tmpreg += 6;  // +10 -16
          tmpreg += 0x90 - ((*dptr)&0xF0);
          if (tmpreg>0x99) tmpreg += 0x60;
       }
        else {
          // binary mode
          tmpreg= - (*dptr);
          //  OVF set is content was 0x80
          if ((*dptr)==0x80) regFlags |= m_OVF;
        }
        // CY set is content was 0
        if ((*dptr)==0) regFlags |= m_CY;
        goto endByte;

      case 0x8e: // NCB
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg= 9 - ((*dptr)&0x0F);
          tmpreg += 0x90 - ((*dptr)&0xF0);
        }
        else {
          // binary mode
          tmpreg= (~(*dptr))&0xff;
        }
        goto endByte;

      case 0x8d: // TCM
        n=lenDR;
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        if (regFlags&m_DCM) {
          // decimal mode (do 100-content)
          tmpreg= 10 - ((*dptr)&0x0F);
          if (tmpreg>9) tmpreg += 6;  // +10 -16
          tmpreg += 0x90 - ((*dptr)&0xF0);
          if (tmpreg>0x99) tmpreg += 0x60;
          if (tmpreg&0x100) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            tmpreg= 9 - ((*dptr)&0x0F);
            if (regFlags & m_CY) tmpreg++;
            if (tmpreg>9) tmpreg += 6;  // +10 -16
            tmpreg += 0x90 - ((*dptr)&0xF0);
            if (tmpreg>0x99) tmpreg += 0x60;  // +10 -16
            if (tmpreg&0x100) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
       }
        else {
          // binary mode
          regFlags |= m_OVF;
          tmpreg= ((~(*dptr))&0xff) + 1;   // bug fix 14/01/17
          if (tmpreg&0x100) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            if (*dptr) regFlags &= ~m_OVF;
            *dptr++ = tmpreg;
            tmpreg= (~(*dptr))&0xff;
            if (regFlags & m_CY) tmpreg++;
            if (tmpreg&0x100) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
          //  OVF set is content was 0x80...0
          if ((*dptr)!=0x80) regFlags &= ~m_OVF;
        }
        goto endMultiple;

      case 0x8f: // NCM
        n=lenDR;
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        if (regFlags&m_DCM) {
          // decimal mode (do 100-content)
          tmpreg= 9 - ((*dptr)&0x0F);
          tmpreg += 0x90 - ((*dptr)&0xF0);
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            tmpreg= 9 - ((*dptr)&0x0F);
            tmpreg += 0x90 - ((*dptr)&0xF0);
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
       }
        else {
          // binary mode
          tmpreg= (~(*dptr))&0xff;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            tmpreg= (~(*dptr))&0xff;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
        }
        goto endMultiple;


      // --------------- test/clear -----------------------

      case 0x90: // TSB
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        tmpreg= *dptr;
        // update Z, MSB, LSB, LDZ, RDZ
        regFlags |= tmpreg & (m_MSB+m_LSB);
        if ((tmpreg&0xFF)==0) regFlags |= m_Z;
        if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        break;

      case 0x91: // TSM
        n=lenDR;
        //  init flags
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        tmpreg=(*dptr);
        // update flags for first byte
        regFlags |= tmpreg & m_LSB;
        if (tmpreg&0xFF) regFlags &= ~m_Z;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        while (n--) {
          dptr++;
          tmpreg= *dptr;
          // update flag Z for each byte
          if (tmpreg&0xFF) regFlags &= ~m_Z;
        }
        // update MSB, LDZ
        regFlags |= tmpreg & m_MSB;
        if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
        break;

      // clear
      case 0x92: // CLB
        *dptr =0;
        regFlags &= ~(m_MSB+m_LSB+m_OVF+m_CY);
        regFlags |= (m_Z+m_LDZ+m_RDZ);
        break;

      case 0x93: // CLM
        n=lenDR;
        do  *dptr++ = 0; while (n--);
        regFlags &= ~(m_MSB+m_LSB+m_OVF+m_CY);
        regFlags |= (m_Z+m_LDZ+m_RDZ);
        break;


      // --------------- or/xor -----------------------

      // or
      case 0x94: // ORB
        sptr= &regCPU.byt[regARP];
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        tmpreg = *dptr | *sptr;
        goto endByte;

      case 0x95: // ORM
        sptr= &regCPU.byt[regARP];
        n=lenDR;
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        tmpreg = *dptr | *sptr;
        // update flags for first byte
        regFlags |= tmpreg & (m_LSB);
        if (tmpreg&0xFF) regFlags &= ~m_Z;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        while (n--) {
          *dptr++ = tmpreg;
          sptr++;
          tmpreg = *dptr | *sptr;
          // update flag Z for each byte
          if (tmpreg&0xFF) regFlags &= ~m_Z;
        }
        goto endMultiple;

      // xor
      case 0x96: // XRB
        sptr= &regCPU.byt[regARP];
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        tmpreg = *dptr ^ *sptr;
        goto endByte;

      case 0x97: // XRM
        sptr= &regCPU.byt[regARP];
        n=lenDR;
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        tmpreg = *dptr ^ *sptr;
        // update flags for first byte
        regFlags |= tmpreg & (m_LSB);
        if (tmpreg&0xFF) regFlags &= ~m_Z;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        while (n--) {
          *dptr++ = tmpreg;
          sptr++;
          tmpreg = *dptr ^ *sptr;
          // update flag Z for each byte
          if (tmpreg&0xFF) regFlags &= ~m_Z;
        }
        goto endMultiple;


      // --------------- inherent 1-byte opcodes -----------------------

      case 0x98: // BIN
        regFlags &= ~m_DCM;
        break;
      case 0x99: // BCD
        regFlags |=  m_DCM;
        break;
      case 0x9a: // SAD
        n=regARP;
        if (regFlags & m_OVF) n |= 0x80;
        if (regFlags & m_CY ) n |= 0x40;
        buf[0]=n;
        n=regDRP;
        if (regFlags & m_OVF) n |= 0x80;
        if (regFlags & m_DCM) n |= 0x40;
        buf[1]=n;
        buf[2] = (regFlags & 0xc7) ^ (m_LDZ+m_Z+m_RDZ) ;
        storeData(regCPU.reg.sp,buf,3);
        regCPU.reg.sp += 3;
        break;
      case 0x9b: // DCE
        regE= (regE-1)&15;
        break;
      case 0x9c: // ICE
        regE= (regE+1)&15;
        break;
      case 0x9d: // CLE
        regE=0;
        break;
      case 0x9e: // RTN
        regCPU.reg.sp -= 2;
        readData(regCPU.reg.sp,buf,2);
        v=buf[0]+(buf[1]<<8);
        regCPU.reg.pc = v-1;   // adjust PC for next increment
        break;
      case 0x9f: // PAD
        regCPU.reg.sp -= 3;
        readData(regCPU.reg.sp,buf,3);
        regARP=buf[0]&0x3f;
        regDRP=buf[1]&0x3f;
        regFlags=buf[2] ^ (m_LDZ+m_Z+m_RDZ);
        if (buf[0] & 0x80) regFlags |= m_OVF;
        if (buf[0] & 0x40) regFlags |= m_CY;
        if (buf[1] & 0x80) regFlags |= m_OVF;
        if (buf[1] & 0x40) regFlags |= m_DCM;
        // need to restore lenDR (fixed 24/2/17)
        if (regDRP&0x20)
          lenDR= 7- (regDRP&7);
        else
          lenDR= 1 - (regDRP&1);
        break;



      // --------------- load -----------------------

      case 0xa8: // LDB  lit imm
        regCPU.reg.pc++;
        tmpreg=fetch1(regCPU.reg.pc);
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        goto endByte;

      case 0xa0: // LDB  reg imm
        tmpreg = regCPU.byt[regARP];
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        goto endByte;

      case 0xa2: // STB  reg imm
        tmpreg = regCPU.byt[regDRP];
        dptr = &regCPU.byt[regARP];  // dptr here means destination
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        goto endByte;

      case 0xa4: // LDBD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
loadB:
        res=readData(v, buf, 1);
        tmpreg= buf[0];
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        goto endByte;

      case 0xb0: // LDBD lit direct
        v=lit_direct();
        goto loadB;

      case 0xb4: // LDBD index direct
        v=index_direct();
        regCPU.reg.idx= v;  // index stored in R02
        goto loadB;

      case 0xac: // LDBI reg ind
        v=reg_ind();
        goto loadB;

      case 0xb8: // LDBI lit ind
        v=lit_ind();
        goto loadB;

      case 0xbc: // LDBI index ind
        v=index_ind();
        goto loadB;

      // pop

      case 0xe0: // POBD dir+
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        // SP+=1
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        goto loadB;

      case 0xe2: // POBD dir-
        // SP-=1
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        goto loadB;

      case 0xe8: // POBI ind+
        v=reg_ind();
        // SP+=2
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        goto loadB;

      case 0xea: // POBI ind-
        // SP-=2
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        v=reg_ind();
        goto loadB;



      // load multiple

      case 0xa9: // LDM  lit imm
        // attn: DRP 4 case (GTO)
        n=lenDR;
        lit_imm(buf,n);  // fill buf with n+1 immediate bytes
        sptr=buf;
        goto loadM;

      case 0xa1: // LDM  reg imm
        sptr = &regCPU.byt[regARP];
loadRegisters:
        n=lenDR;
        goto loadM;

      case 0xa3: // STM  reg imm
        sptr = &regCPU.byt[regDRP];
        dptr = &regCPU.byt[regARP];
        goto loadRegisters;

      case 0xa5: // LDMD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
loadMultiple:
        n=lenDR;
        res=readData(v, buf, n+1);
        sptr= buf;
loadM:
        //  init flags
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        tmpreg= *sptr++;
        // update flags for first byte
        regFlags |= tmpreg & (m_LSB);
        if (tmpreg&0xFF) regFlags &= ~m_Z;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        while (n--) {
          *dptr++ = tmpreg;
          tmpreg= *sptr++;
          // update flag Z for each byte
          if (tmpreg&0xFF) regFlags &= ~m_Z;
        }
        goto endMultiple;

      case 0xb1: // LDMD lit direct
        v=lit_direct();
        goto loadMultiple;

      case 0xb5: // LDMD index direct
        v=index_direct();
        regCPU.reg.idx= v;  // index stored in R02
        goto loadMultiple;

      case 0xad: // LDMI reg ind
        v=reg_ind();
        goto loadMultiple;

      case 0xb9: // LDMI lit ind
        v=lit_ind();
        goto loadMultiple;

      case 0xbd: // LDMI index ind
        v=index_ind();
        goto loadMultiple;


      // pop
      case 0xe1: // POMD dir+
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        n=lenDR;
        // SP+=n
        i=regCPU.byt[regARP];
        regCPU.byt[regARP] += n+1;
        if (regCPU.byt[regARP]<i) regCPU.byt[regARP+1]++;
        goto loadMultiple;

      case 0xe3: // POMD dir-
        n=lenDR;
        // SP-=n
        i=regCPU.byt[regARP];
        regCPU.byt[regARP] -= n+1;
        if (regCPU.byt[regARP]>i) regCPU.byt[regARP+1]--;
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        goto loadMultiple;

      case 0xe9: // POMI ind+
        v=reg_ind();
        // SP+=2
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        goto loadMultiple;

      case 0xeb: // POMI ind-
        // SP-=2
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        v=reg_ind();
        goto loadMultiple;



      // --------------- store -----------------------

      case 0xa6: // STBD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
storeByte:
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        tmpreg= regCPU.byt[regDRP];
        buf[0]= tmpreg;
        // update Z, MSB, LSB, LDZ, RDZ
        regFlags |= tmpreg & (m_MSB+m_LSB);
        if ((tmpreg&0xFF)==0) regFlags |= m_Z;
        if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        res=storeData(v,buf,1);
        break;

      case 0xb2: // STBD lit direct
        v=lit_direct();
        goto storeByte;

      case 0xae: // STBI reg ind
        v=reg_ind();
        goto storeByte;

      case 0xb6: // STBD index direct
        v=index_direct();
        regCPU.reg.idx= v;  // index stored in R02
        goto storeByte;

      case 0xba: // STBI lit ind
        v=lit_ind();
        goto storeByte;

      case 0xbe: // STBI index ind
        v=index_ind();
        goto storeByte;

      // push

      case 0xe4: // PUBD dir+
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        // SP+=1
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        // that may look strange to increment SP here before actual move,
        // but it seems that it works like that (see PUBD R2,+R2 in UNROM)
        goto storeByte;

      case 0xe6: // PUBD dir-
        // SP-=1
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        goto storeByte;

      case 0xec: // PUBI ind+
        v=reg_ind();
        // SP+=2
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        goto storeByte;

      case 0xee: // PUBI ind-
        // SP-=2
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        v=reg_ind();
        goto storeByte;



      // store multiple

      case 0xa7: // STMD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
storeMultiple:
        sptr= &regCPU.byt[regDRP];
        dptr = buf;
        n=lenDR;
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        tmpreg= *sptr++;
        // update flags for first byte
        regFlags |= tmpreg & (m_LSB);
        if (tmpreg&0xFF) regFlags &= ~m_Z;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        i=n;
        while (i--) {
          *dptr++ = tmpreg;
          tmpreg= *sptr++;
          // update flag Z for each byte
          if (tmpreg&0xFF) regFlags &= ~m_Z;
        }
        *dptr = tmpreg;
        // update MSB, LDZ
        regFlags |= tmpreg & (m_MSB);
        if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
        res=storeData(v,buf,n+1);
        break;

      case 0xb3: // STMD lit direct
        v=lit_direct();
        goto storeMultiple;

      case 0xaf: // STMI reg ind
        v=reg_ind();
        goto storeMultiple;

      case 0xb7: // STMD index direct
        v=index_direct();
        regCPU.reg.idx= v;  // index stored in R02
        goto storeMultiple;

      case 0xbb: // STMI lit ind
        v=lit_ind();
        goto storeMultiple;

      case 0xbf: // STMI index ind
        v=index_ind();
        goto storeMultiple;


      case 0xaa: // STB  lit imm ?
      case 0xab: // STM  lit imm ?
        res=-1;
        break;

      // push

      case 0xe5: // PUMD dir+
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        n=lenDR;
        // SP+=n
        i=regCPU.byt[regARP];
        regCPU.byt[regARP] += n+1;
        if (regCPU.byt[regARP]<i) regCPU.byt[regARP+1]++;
        goto storeMultiple;

      case 0xe7: // PUMD dir-
        n=lenDR;
        // SP-=n
        i=regCPU.byt[regARP];
        regCPU.byt[regARP] -= n+1;
        if (regCPU.byt[regARP]>i) regCPU.byt[regARP+1]--;
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        goto storeMultiple;

      case 0xed: // PUMI ind+
        v=reg_ind();
        // SP+=2
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        if ((++regCPU.byt[regARP])==0) regCPU.byt[regARP+1]++;
        goto storeMultiple;

      case 0xef: // PUMI ind-
        // SP-=2
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        if ((regCPU.byt[regARP]--)==0) regCPU.byt[regARP+1]--;
        v=reg_ind();
        goto storeMultiple;


      // --------------- add -----------------------

      case 0xc2: // ADB  reg imm
        sptr= &regCPU.byt[regARP];
addByte:
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)+((*sptr)&0x0F);
          if (tmpreg>0x09) tmpreg += 6; // -10 + 16
          tmpreg += ((*dptr)&0xF0)+((*sptr)&0xF0);
          if (tmpreg>0x99) tmpreg += 0x60;  // -10*16
        }
        else {
          // binary mode
          tmpreg = *dptr + *sptr;
          // OVF if src>0, dest>0 and res<0
          //  or if src<0, dest<0 and res>0
          // i.e. OVF if sgn(src)=sgn(dest) and sgn(src)<>sgn(res)
          if ( (((*sptr)&0x80)==((*dptr)&0x80)) && (((*sptr)&0x80)!=(tmpreg&0x80)) )
            regFlags |= m_OVF;
        }
        if (tmpreg&0x100) regFlags |= m_CY;
        goto endByte;

      case 0xca: // ADB  lit imm
        regCPU.reg.pc++;
        buf[0]=fetch1(regCPU.reg.pc);
        sptr = buf;
        goto addByte;

      case 0xd2: // ADB  lit direct
        v=lit_direct();
addB:
        readData(v, buf, 1);
        sptr= buf;
        goto addByte;

      case 0xda: // ADBD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        goto addB;

      // multiple

      case 0xc3: // ADM  reg imm
        sptr= &regCPU.byt[regARP];
        n=lenDR;
addM:
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)+((*sptr)&0x0F);
          if (tmpreg>0x09) tmpreg += 6; // -10 + 16
          tmpreg += ((*dptr)&0xF0)+((*sptr)&0xF0);
          if (tmpreg>0x99) tmpreg += 0x60;  // -10*16
          if (tmpreg&0x100) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            sptr++;
            tmpreg=((*dptr)&0x0F)+((*sptr)&0x0F);
            if (regFlags & m_CY) tmpreg++;
            if (tmpreg>0x09) tmpreg += 6; // -10 + 16
            tmpreg += ((*dptr)&0xF0)+((*sptr)&0xF0);
            if (tmpreg>0x99) tmpreg += 0x60;  // -10*16
            if (tmpreg&0x100) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
        }
        else {
          // binary mode
          tmpreg = *dptr + *sptr;
          if (tmpreg&0x100) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            sptr++;
            tmpreg = *dptr + *sptr;
            if (regFlags & m_CY) tmpreg++;
            if (tmpreg&0x100) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
          // update OVF
          if ( (((*sptr)&0x80)==((*dptr)&0x80)) && (((*sptr)&0x80)!=(tmpreg&0x80)) )
            regFlags |= m_OVF;
        }
        goto endMultiple;

      case 0xd3: // ADM  lit direct
        v=lit_direct();
        n=lenDR;
        readData(v,buf,n+1);
        sptr=buf;
        goto addM;

      case 0xdb: // ADMD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        n=lenDR;
        readData(v,buf,n+1);
        sptr=buf;
        goto addM;

      case 0xcb: // ADM  lit imm
        n=lenDR;
        lit_imm(buf,n);
        sptr=buf;
        goto addM;

      // --------------- sub -----------------------

      case 0xc4: // SBB  reg imm
        sptr= &regCPU.byt[regARP];
subByte:
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)-((*sptr)&0x0F);
          if (tmpreg<0) tmpreg -= 6;
          tmpreg += ((*dptr)&0xF0)-((*sptr)&0xF0);
          if (tmpreg<0) tmpreg -= 0x60;
        }
        else {
          // binary mode
          tmpreg = *dptr - *sptr;
          // OVF if dest>0, src<0 and res<0
          //  or if dest<0, src>0 and res>0
          // i.e. OVF if sgn(src)<>sgn(dest) and sgn(dest)<>sgn(res)
          if ( (((*dptr)&0x80)!=((*sptr)&0x80)) && (((*dptr)&0x80)!=(tmpreg&0x80)) )
            regFlags |= m_OVF;
        }
        if (tmpreg>=0) regFlags |= m_CY;
        goto endByte;

      case 0xcc: // SBB  lit imm
        regCPU.reg.pc++;
        buf[0]=fetch1(regCPU.reg.pc);
        sptr = buf;
        goto subByte;

      case 0xd4: // SBB  lit direct
        v=lit_direct();
subB:
        readData(v, buf, 1);
        sptr= buf;
        goto subByte;

      case 0xdc: // SBBD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        goto subB;


      // multiple

      case 0xc5: // SBM  reg imm
        sptr= &regCPU.byt[regARP];
        n=lenDR;
subM:
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)-((*sptr)&0x0F);
          if (tmpreg<0) tmpreg -= 6;
          tmpreg += ((*dptr)&0xF0)-((*sptr)&0xF0);
          if (tmpreg<0) tmpreg -= 0x60;  // -10*16
          if (tmpreg<0) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            sptr++;
            tmpreg=((*dptr)&0x0F)-((*sptr)&0x0F);
            if (regFlags & m_CY) tmpreg--;
            if (tmpreg<0) tmpreg -= 6;
            tmpreg += ((*dptr)&0xF0)-((*sptr)&0xF0);
            if (tmpreg<0) tmpreg -= 0x60;
            if (tmpreg<0) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
        }
        else {
          // binary mode
          tmpreg = *dptr - *sptr;
          if (tmpreg&0x100) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            *dptr++ = tmpreg;
            sptr++;
            tmpreg = *dptr - *sptr;
            if (regFlags & m_CY) tmpreg--;
            if (tmpreg<0) regFlags |= m_CY; else regFlags &= ~m_CY; // !
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
          // update OVF
          if ( (((*dptr)&0x80)!=((*sptr)&0x80)) && (((*dptr)&0x80)!=(tmpreg&0x80)) )
            regFlags |= m_OVF;
        }
        regFlags ^= m_CY;
        goto endMultiple;

      case 0xd5: // SBM  lit direct
        v=lit_direct();
        n=lenDR;
        readData(v,buf,n+1);
        sptr=buf;
        goto subM;

      case 0xdd: // SBMD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        n=lenDR;
        readData(v,buf,n+1);
        sptr=buf;
        goto subM;

      case 0xcd: // SBM  lit imm
        n=lenDR;
        lit_imm(buf,n);
        sptr=buf;
        goto subM;

      // --------------- compare -----------------------

      case 0xc0: // CMB  reg imm
        sptr= &regCPU.byt[regARP];
cmpByte:
        regFlags &= ~(m_Z+m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)-((*sptr)&0x0F);
          if (tmpreg<0) tmpreg -= 6;
          tmpreg += ((*dptr)&0xF0)-((*sptr)&0xF0);
          if (tmpreg<0) tmpreg -= 0x60;
        }
        else {
          // binary mode
          tmpreg = *dptr - *sptr;
          // OVF
          if ( (((*dptr)&0x80)!=((*sptr)&0x80)) && (((*dptr)&0x80)!=(tmpreg&0x80)) )
            regFlags |= m_OVF;
        }
        if (tmpreg>=0) regFlags |= m_CY;   // fixed 15/2/17
        // update Z, MSB, LSB, LDZ, RDZ
        regFlags |= tmpreg & (m_MSB+m_LSB);
        if ((tmpreg&0xFF)==0) regFlags |= m_Z;
        if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        break;

      case 0xc8: // CMB  lit imm
        regCPU.reg.pc++;
        buf[0]=fetch1(regCPU.reg.pc);
        sptr = buf;
        goto cmpByte;

      case 0xd0: // CMB  lit direct
        v=lit_direct();
cmpB:
        readData(v, buf, 1);
        sptr= buf;
        goto cmpByte;

      case 0xd8: // CMBD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        goto cmpB;

      // multiple

      case 0xc1: // CMM  reg imm
        sptr= &regCPU.byt[regARP];
        n=lenDR;
cmpM:
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        if (regFlags&m_DCM) {
          // decimal mode
          tmpreg=((*dptr)&0x0F)-((*sptr)&0x0F);
          if (tmpreg<0) tmpreg -= 6;
          tmpreg += ((*dptr)&0xF0)-((*sptr)&0xF0);
          if (tmpreg<0) tmpreg -= 0x60;  // -10*16
          if (tmpreg<0) regFlags |= m_CY;
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            dptr++;
            sptr++;
            tmpreg=((*dptr)&0x0F)-((*sptr)&0x0F);
            if (regFlags & m_CY) tmpreg--;
            if (tmpreg<0) tmpreg -= 6;
            tmpreg += ((*dptr)&0xF0)-((*sptr)&0xF0);
            if (tmpreg<0) tmpreg -= 0x60;
            if (tmpreg<0) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
        }
        else {
          // binary mode
          tmpreg = *dptr - *sptr;
          if (tmpreg<0) regFlags |= m_CY; // !
          // update flags for first byte
          regFlags |= tmpreg & (m_LSB);
          if (tmpreg&0xFF) regFlags &= ~m_Z;
          if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
          while (n--) {
            dptr++;
            sptr++;
            tmpreg = *dptr - *sptr;
            if (regFlags & m_CY) tmpreg--;
            if (tmpreg<0) regFlags |= m_CY; else regFlags &= ~m_CY;
            // update flag Z for each byte
            if (tmpreg&0xFF) regFlags &= ~m_Z;
          }
          // update OVF
          if ( (((*dptr)&0x80)!=((*sptr)&0x80)) && (((*dptr)&0x80)!=(tmpreg&0x80)) )
            regFlags |= m_OVF;
        }
        regFlags ^= m_CY;
        // update MSB, LDZ
        regFlags |= tmpreg & (m_MSB);
        if ((tmpreg&0xF0)==0) regFlags |= m_LDZ;
        break;

      case 0xd1: // CMM  lit direct
        v=lit_direct();
        n=lenDR;
        readData(v,buf,n+1);
        sptr=buf;
        goto cmpM;

      case 0xd9: // CMMD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        n=lenDR;
        readData(v,buf,n+1);
        sptr=buf;
        goto cmpM;

      case 0xc9: // CMM  lit imm
        n=lenDR;
        lit_imm(buf,n);
        sptr=buf;
        goto cmpM;


      // --------------- and -----------------------

      case 0xc7: // ANM  reg imm
        sptr= &regCPU.byt[regARP];
        n=lenDR;
andM:
        regFlags &= ~(m_MSB+m_LDZ+m_RDZ+m_LSB+m_OVF+m_CY);
        regFlags |= m_Z;
        // binary mode
        tmpreg = *dptr & *sptr;
        // update flags for first byte
        regFlags |= tmpreg & (m_LSB);
        if (tmpreg&0xFF) regFlags &= ~m_Z;
        if ((tmpreg&0x0F)==0) regFlags |= m_RDZ;
        while (n--) {
          *dptr++ = tmpreg;
          sptr++;
          tmpreg = *dptr & *sptr;
          // update flag Z for each byte
          if (tmpreg&0xFF) regFlags &= ~m_Z;
        }
        goto endMultiple;


      case 0xd7: // ANM  lit direct
        v=lit_direct();
        n=lenDR;
        readData(v,buf,n+1);
        sptr=buf;
        goto andM;

      case 0xdf: // ANMD reg direct
        v = regCPU.byt[regARP]+(regCPU.byt[regARP+1]<<8);
        n=lenDR;
        readData(v,buf,n+1);
        sptr=buf;
        goto andM;

      case 0xcf: // ANM  lit imm
        n=lenDR;
        lit_imm(buf,n);
        sptr=buf;
        goto andM;

      // --------------- jsb -----------------------

      case 0xce: // JSB  lit direct
        v=lit_direct();
dojsb:
        // push PC
        i=regCPU.reg.pc + 1; // save the addr after the JSB
        buf[0]=i; buf[1]=i>>8;
        storeData(regCPU.reg.sp,buf,2);
        // SP +=2
        regCPU.reg.sp += 2;
        // jump
        regCPU.reg.pc = v-1;  // adjust PC for next increment
        break;

      case 0xc6: // JSB  index direct
        v=index_direct();
        goto dojsb;


      // --------------- nop -----------------------

      case 0xd6: // NOP3
        regCPU.reg.pc += 2;
        break;

      case 0xde: // NOP1
        res=-2;   // breakpoint
        break;


      // --------------- push/pop -----------------------
      // done in load/store


      // --------------- jmp -----------------------

      case 0xf0: // JMP
        f=1;
jmprel:
        regCPU.reg.pc++;
        if (f) {
          v= fetch1(regCPU.reg.pc);
          if (v>127) v -= 256;
          regCPU.reg.pc += v;
        }
        break;
      case 0xf1: // JNO
        f=!(regFlags&m_OVF); goto jmprel;
      case 0xf2: // JOD
        f= (regFlags&m_LSB); goto jmprel;
      case 0xf3: // JEV
        f=!(regFlags&m_LSB); goto jmprel;
      case 0xf4: // JNG    jump on negative
        if (regFlags&m_OVF)
          f=!(regFlags&m_MSB);
        else
          f=(regFlags&m_MSB);
        goto jmprel;
      case 0xf5: // JPS jump on positive
        if (regFlags&m_OVF)
          f=(regFlags&m_MSB);
        else
          f=!(regFlags&m_MSB);
        goto jmprel;
      case 0xf6: // JNZ
        f=!(regFlags&m_Z  ); goto jmprel;
      case 0xf7: // JZR
        f= (regFlags&m_Z  ); goto jmprel;
      case 0xf8: // JEN
        f=regE; ; goto jmprel;
      case 0xf9: // JEZ
        f=(regE==0); ; goto jmprel;
      case 0xfa: // JNC
        f=!(regFlags&m_CY ); goto jmprel;
      case 0xfb: // JCY
        f= (regFlags&m_CY ); goto jmprel;
      case 0xfc: // JLZ
        f= (regFlags&m_LDZ); goto jmprel;
      case 0xfd: // JLN
        f=!(regFlags&m_LDZ); goto jmprel;
      case 0xfe: // JRZ
        f= (regFlags&m_RDZ); goto jmprel;
      case 0xff: // JRN
        f=!(regFlags&m_RDZ); goto jmprel;

      default:
        res=-1;


    } // end switch

    // increment PC, now
    regCPU.reg.pc++;

    if (res) {
      if (res<0) regCPU.reg.pc--;  // restore pc to the break point
      break;
    }


  } // end while

  return(res);

}




